# SPDX-FileCopyrightText: 2026-present Vaclav Dvorak <vashek@gmail.com>
#
# SPDX-License-Identifier: MIT
import tempfile
from pathlib import Path
from typing import ClassVar

import pytest  # type: ignore

import save_as_xlsx

from .test_pyopenxl_verifier import verify_using_pyopenxl

try:
    import pydantic  # type: ignore
    BaseModel: type = pydantic.BaseModel
except ImportError:
    pydantic = None
    class BaseModel:  # type: ignore
        def __init__(self, *args, **kwargs):
            pass


def test_print_pydantic_version():
    print("pydantic version: " + getattr(pydantic, "__version__", "(none)") + "\n")


@pytest.mark.skipif(pydantic is None, reason="requires pydantic")
class TestPydantic:
    class ModelForTest(BaseModel):
        a: int
        b: str | None = None

    TEST_DATA_WITH_PYDANTIC: ClassVar[list[ModelForTest]] = [
        ModelForTest(a=1),
        ModelForTest(a=2, b="B"),
    ]

    def test_save_pydantic(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            fn = Path(tmpdir) / "test.xlsx"
            with save_as_xlsx.SaveAsXlsx(fn, self.TEST_DATA_WITH_PYDANTIC) as saver:
                assert len(saver.columns_values) == 2
                assert saver.columns_values[0]["header"] == "a"
                assert saver.columns_values[1]["header"] == "b"
                assert saver.number_of_value_rows == 2
            assert fn.exists()
            verify_using_pyopenxl(fn, "A1:B3", data=[
                ("a", "b"),
                (1, None),
                (2, "B"),
            ])



# SPDX-FileCopyrightText: 2026-present Vaclav Dvorak <vashek@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import json
from collections.abc import Iterable, Mapping, Set, Sized
from dataclasses import asdict, fields, is_dataclass
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum, IntEnum
from fractions import Fraction
from os import PathLike, fspath
from typing import Annotated, Any, ClassVar, Protocol, TypeAlias
from uuid import UUID

import xlsxwriter  # type: ignore
import xlsxwriter.worksheet  # type: ignore
from annotated_types import Gt, Unit, Lt
from xlsxwriter.exceptions import XlsxWriterException  # type: ignore

try:
    import pydantic  # type: ignore
    from pydantic import BaseModel
    PYDANTIC_VER = int(pydantic.__version__.split(".")[0])
except ImportError:
    class BaseModel: pass  # type: ignore
    PYDANTIC_VER = -1

from .__about__ import __version__


__all__ = [
    "__version__",
    "SaveAsXlsx",
    "save_as_xlsx",
    "ColumnWidth",
    "TableAddError",
    "WorkbookClosedError",
    "UnsupportedTypeError",
]


class TableAddError(XlsxWriterException):
    pass


class WorkbookClosedError(XlsxWriterException):
    pass


class UnsupportedTypeError(XlsxWriterException, TypeError):
    pass


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Any]]


class ColumnWidth(IntEnum):
    AUTOFIT = -1
    HIDE = 0


ColumnWidthType: TypeAlias = None | ColumnWidth | Annotated[int|float, Gt(0), Unit("px")] | Annotated[int|float, Lt(0), Unit("characterUnits")]


class SaveAsXlsx:
    def __init__(self,
                 filename: str | PathLike,
                 data: Iterable[Mapping[str, Any] | DataclassInstance | BaseModel] | None = None,
                 sheet_name: str | None = None,
                 table_name: str | None = None,
                 column_order: Iterable[str] | None = None,
                 column_width: ColumnWidthType | Mapping[str, ColumnWidthType] | Iterable[ColumnWidthType] = None,
                 *,
                 extra_columns: bool = True,
                 total_row: bool = False,
                 auto_save: bool = False,
                 ) -> None:
        self.closed = False
        self.filename = filename
        self.workbook = xlsxwriter.Workbook(fspath(filename))
        self.worksheet: xlsxwriter.worksheet.Worksheet | None = None
        self.columns: dict[str, dict[str, str | int | float]] = {}
        self.columns_values: tuple[dict[str, str | int | float], ...] = ()
        self.number_of_value_rows = 0
        if data is not None:
            self.add_sheet(data, sheet_name=sheet_name, table_name=table_name, column_order=column_order, column_width=column_width, extra_columns=extra_columns, total_row=total_row)
        if auto_save:
            self.close()

    def add_sheet(self,
                  data: Iterable[Mapping[str, Any] | DataclassInstance | BaseModel],
                  sheet_name: str | None = None,
                  table_name: str | None = None,
                  column_order: Iterable[str] | None = None,
                  column_width: ColumnWidthType | Mapping[str, ColumnWidthType] | Iterable[ColumnWidthType] = None,
                  *,
                  extra_columns: bool = True,
                  total_row: bool = False,
                  ) -> xlsxwriter.worksheet.Worksheet:
        if self.closed:
            raise WorkbookClosedError()
        self.worksheet = worksheet = self.workbook.add_worksheet(sheet_name)
        columns: dict[str, dict[str, str | int | float]] = {}
        self.columns = columns
        for column in column_order or ():
            columns[column] = {"header": column}
        if not isinstance(data, Iterable):
            raise TypeError("data must be an iterable")
        if not isinstance(data, Sized):
            data = tuple(data)
        if extra_columns:
            for row in data:
                # missing_cols = row.keys() - columns.keys()  # not order-preserving, so instead:
                missing_cols = [col_name for col_name in (
                    (f.name for f in fields(row)) if is_dataclass(row) else
                    type(row).model_fields.keys() if PYDANTIC_VER >= 2 and isinstance(row, BaseModel) else
                    row.__fields__.keys() if isinstance(row, BaseModel) else
                    row.keys()  # type: ignore
                ) if col_name not in columns]
                for column in missing_cols:
                    columns[column] = {"header": column}
        col_names = columns.keys()
        self.columns_values = tuple(columns.values())
        self.number_of_value_rows = len(data)
        result = worksheet.add_table(0, 0, len(data), len(columns) - 1, {
            "header_row": True,
            "columns": self.columns_values,
            "total_row": total_row,
            **({"name": table_name} if table_name else {}),
            "data": [
                [self.convert_value(row_dict.get(col_name)) for col_name in col_names]  # type: ignore
                for row_union in data
                if (row_dict := (asdict(row_union) if is_dataclass(row_union) else  # type: ignore
                                 row_union.model_dump() if PYDANTIC_VER >= 2 and isinstance(row_union, BaseModel) else
                                 row_union.dict() if isinstance(row_union, BaseModel) else
                                 row_union))
            ],
        })
        if result != 0:
            raise TableAddError(f"Table add error: {result}")
        self.set_column_widths(worksheet, columns, column_width)
        return worksheet

    @classmethod
    def set_column_widths(cls,
                          worksheet: xlsxwriter.worksheet.Worksheet,
                          columns: dict[str, dict[str, str | int | float]],
                          column_width: ColumnWidthType | Mapping[str, ColumnWidthType] | Iterable[ColumnWidthType],
                          ) -> None:
        if column_width is None:
            return
        if column_width == ColumnWidth.AUTOFIT:
            worksheet.autofit()
        elif column_width == ColumnWidth.HIDE:
            worksheet.hide()
        elif isinstance(column_width, (int, float)):
            if column_width == 0:
                raise ValueError("column_width cannot be 0, use None to keep the default width or ColumnWidth.AUTOFIT or ColumnWidth.HIDE")
            if column_width < 0:
                worksheet.set_column(0, len(columns) - 1, -column_width)
            else:
                worksheet.set_column_pixels(0, len(columns) - 1, column_width)
        elif isinstance(column_width, Mapping):
            column_keys = tuple(columns.keys())
            for column_name_or_number, desired_width in column_width.items():
                try:
                    column_number = column_keys.index(column_name_or_number)
                except ValueError:
                    if isinstance(column_name_or_number, int) and column_name_or_number >= 0:
                        column_number = column_name_or_number
                    else:
                        raise
                cls.set_column_width(worksheet, column_number, desired_width)
        elif isinstance(column_width, Iterable):
            for column_number, desired_width in enumerate(column_width):
                cls.set_column_width(worksheet, column_number, desired_width)
        else:
            raise TypeError(f"unsupported column_width type: {column_width!r}")

    @staticmethod
    def set_column_width(worksheet: xlsxwriter.worksheet.Worksheet, column_number: int, column_width: ColumnWidthType,
                         ) -> None:
        if column_width is None:
            return
        if column_width == ColumnWidth.AUTOFIT:
            # dangerous, fragile hack...
            orig_dim_colmin, orig_dim_colmax = worksheet.dim_colmin, worksheet.dim_colmax
            worksheet.dim_colmin, worksheet.dim_colmax = column_number, column_number
            worksheet.autofit()
            worksheet.dim_colmin, worksheet.dim_colmax = orig_dim_colmin, orig_dim_colmax
        elif column_width == ColumnWidth.HIDE:
            worksheet.set_column(column_number, column_number, 8.43, None, {'hidden': 1})
        elif isinstance(column_width, (int, float)):
            if column_width == 0:
                raise ValueError("column_width cannot be 0, use None to keep the default width or ColumnWidth.AUTOFIT or ColumnWidth.HIDE")
            if column_width < 0:
                worksheet.set_column(column_number, column_number, -column_width)
            else:
                worksheet.set_column_pixels(column_number, column_number, column_width)
        else:
            raise TypeError(f"unsupported column_width type: {column_width!r}")

    def close(self) -> None:
        if not self.closed:
            self.workbook.close()
        self.closed = True

    def __enter__(self) -> SaveAsXlsx:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type is None and exc_val is None:
            self.close()

    @classmethod
    def convert_value(cls, input_value, *, for_json: bool = False):
        if isinstance(input_value, Enum):  # must be first, because enum may match str or int
            return input_value.name
        if isinstance(input_value, (str, int, float, bool, Decimal, Fraction, datetime, date, time, timedelta)):
            if for_json and isinstance(input_value, (Decimal, Fraction)):
                return float(input_value)
            return input_value
        if input_value is None:
            return None
        if isinstance(input_value, UUID):
            return str(input_value)
        if isinstance(input_value, Mapping):
            return json.dumps(input_value, default=lambda value: cls.convert_value(value, for_json=True))
        if isinstance(input_value, Set):
            return "{" + ", ".join(str(cls.convert_value(value)) for value in input_value) + "}"
        if isinstance(input_value, Iterable):
            return "[" + ", ".join(str(cls.convert_value(value)) for value in input_value) + "]"
        raise UnsupportedTypeError(input_value)


def save_as_xlsx(filename: str | PathLike,
                 data: Iterable[Mapping[str, Any] | DataclassInstance | BaseModel] | None = None,
                 **kwargs) -> None:
    if "auto_save" in kwargs and kwargs["auto_save"] is not None and not kwargs["auto_save"]:
        raise ValueError("calling save_as_xlsx(auto_save=False) makes no sense")
    kwargs["auto_save"] = True
    SaveAsXlsx(filename, data, **kwargs)



# SPDX-FileCopyrightText: 2026-present Vaclav Dvorak <vashek@gmail.com>
#
# SPDX-License-Identifier: MIT
import os.path
import tempfile
from dataclasses import dataclass
from decimal import Decimal
from enum import IntEnum
from fractions import Fraction
from pathlib import Path
from uuid import UUID

import save_as_xlsx

from .test_pyopenxl_verifier import verify_using_pyopenxl


class EnumForTest(IntEnum):
    ONE = 1
    TWO = 2


@dataclass
class DataclassForTest:
    a: int
    b: str | None = None


TEST_DATA = [
    {"a": 1, "b": "2"},
    {"b": "B", "c": None},
]

TEST_DATA_COMPLEX = [
    {"dict": {"a": 1, "b": "2"}},
    {"list": [1, 2, 3]},
    {"set": {1, 2, 3}},
    {"float": -1.5, "dec": Decimal("2.99"), "frac": Fraction(3, 5), "bool": True},
    {"dict": {"nested": {"enum": EnumForTest.ONE, "dec": Decimal("2.99")}}},
]

TEST_DATA_WITH_ENUM = [
    {"a": 1, "enum": EnumForTest.ONE},
]

TEST_DATA_WITH_UUID = [
    {"a": 1, "uuid": UUID("5f456a18-29f0-11f1-a203-e41fd5b9abcb")},
]

TEST_DATA_WITH_DATACLASS = [
    DataclassForTest(a=1),
    DataclassForTest(a=2, b="B"),
]


def test_save_with_function():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        assert not fn.exists()
        save_as_xlsx.save_as_xlsx(fn, TEST_DATA)
        assert fn.exists()
        verify_using_pyopenxl(fn, "A1:C3")

def test_save_on_explicit_close():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        saver = save_as_xlsx.SaveAsXlsx(fn, TEST_DATA)
        assert not fn.exists()
        saver.close()
        assert fn.exists()
        verify_using_pyopenxl(fn, "A1:C3")

def test_save_with_str_filename():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = os.path.join(tmpdir, "test.xlsx")
        assert isinstance(fn, str)
        saver = save_as_xlsx.SaveAsXlsx(fn, TEST_DATA)
        assert not Path(fn).exists()
        saver.close()
        assert Path(fn).exists()
        verify_using_pyopenxl(fn, "A1:C3")

def test_save_on_auto_close():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, auto_save=True)
        assert fn.exists()
        verify_using_pyopenxl(fn, "A1:C3")

def test_save_on_with():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA):
            assert not fn.exists()
        assert fn.exists()
        verify_using_pyopenxl(fn, "A1:C3")

def test_num_rows_and_sheet_name():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA) as saver:
            assert saver.number_of_value_rows == 2
            assert saver.worksheet.name == "Sheet1"
        verify_using_pyopenxl(fn, "A1:C3")

def test_custom_sheet_name():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, sheet_name="My Sheet") as saver:
            assert saver.worksheet.name == "My Sheet"
        verify_using_pyopenxl(fn, "A1:C3", sheet_name="My Sheet")

def test_default_column_order():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA) as saver:
            assert len(saver.columns_values) == 3
            assert saver.columns_values[0]["header"] == "a"
            assert saver.columns_values[1]["header"] == "b"
            assert saver.columns_values[2]["header"] == "c"
        verify_using_pyopenxl(fn, "A1:C3", data=[
            ("a", "b", "c"),
            (1, "2", None),
            (None, "B", None),
        ])

def test_column_order():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, column_order=("b",)) as saver:
            assert len(saver.columns_values) == 3
            assert saver.columns_values[0]["header"] == "b"
            assert saver.columns_values[1]["header"] == "a"
            assert saver.columns_values[2]["header"] == "c"
        verify_using_pyopenxl(fn, "A1:C3", data=[
            ("b", "a", "c"),
            ("2", 1, None),
            ("B", None, None),
        ])

def test_column_order_extra_nonexistent_column():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, column_order=("b", "nonexistent")) as saver:
            assert len(saver.columns_values) == 4
            assert saver.columns_values[0]["header"] == "b"
            assert saver.columns_values[1]["header"] == "nonexistent"
            assert saver.columns_values[2]["header"] == "a"
            assert saver.columns_values[3]["header"] == "c"
        verify_using_pyopenxl(fn, "A1:D3", data=[
            ("b", "nonexistent", "a", "c"),
            ("2", None, 1, None),
            ("B", None, None, None),
        ])

def test_column_order_no_extras():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, column_order=("b", "c"), extra_columns=False) as saver:
            assert len(saver.columns_values) == 2
            assert saver.columns_values[0]["header"] == "b"
            assert saver.columns_values[1]["header"] == "c"
        verify_using_pyopenxl(fn, "A1:B3", data=[
            ("b", "c"),
            ("2", None),
            ("B", None),
        ])

def test_column_order_extra_nonexistent_column_no_extras():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, column_order=("b", "nonexistent"), extra_columns=False) as saver:
            assert len(saver.columns_values) == 2
            assert saver.columns_values[0]["header"] == "b"
            assert saver.columns_values[1]["header"] == "nonexistent"
        verify_using_pyopenxl(fn, "A1:B3", data=[
            ("b", "nonexistent"),
            ("2", None),
            ("B", None),
        ])

def test_generator():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, ({"num": i} for i in range(5))) as saver:
            assert len(saver.columns_values) == 1
            assert saver.columns_values[0]["header"] == "num"
            assert saver.number_of_value_rows == 5
        verify_using_pyopenxl(fn, "A1:A6", data=[
            ("num",),
            (0,),
            (1,),
            (2,),
            (3,),
            (4,),
        ])

def test_enum():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA_WITH_ENUM) as saver:
            assert len(saver.columns_values) == 2
            assert saver.columns_values[0]["header"] == "a"
            assert saver.columns_values[1]["header"] == "enum"
        verify_using_pyopenxl(fn, "A1:B2", data=[
            ("a", "enum"),
            (1, "ONE"),
        ])

def test_uuid():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA_WITH_UUID) as saver:
            assert len(saver.columns_values) == 2
            assert saver.columns_values[0]["header"] == "a"
            assert saver.columns_values[1]["header"] == "uuid"
        verify_using_pyopenxl(fn, "A1:B2", data=[
            ("a", "uuid"),
            (1, "5f456a18-29f0-11f1-a203-e41fd5b9abcb"),
        ])

def test_complex():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA_COMPLEX) as saver:
            assert len(saver.columns_values) == 7
            assert saver.number_of_value_rows == len(TEST_DATA_COMPLEX)
        verify_using_pyopenxl(fn, "A1:G6", data=[
            ("dict", "list", "set", "float", "dec", "frac", "bool"),
            ('{"a": 1, "b": "2"}', None, None, None, None, None, None),
            (None, '[1, 2, 3]', None, None, None, None, None),
            (None, None, '{1, 2, 3}', None, None, None, None),
            (None, None, None, -1.5, 2.99, 0.6, True),
            # TODO: XXX BUG enum should be "ONE"
            ('{"nested": {"enum": 1, "dec": 2.99}}', None, None, None, None, None, None),
        ])

def test_dataclasses():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA_WITH_DATACLASS) as saver:
            assert len(saver.columns_values) == 2
            assert saver.columns_values[0]["header"] == "a"
            assert saver.columns_values[1]["header"] == "b"
            assert saver.number_of_value_rows == 2
        verify_using_pyopenxl(fn, "A1:B3", data=[
            ("a", "b"),
            (1, None),
            (2, "B"),
        ])

def test_another_sheet():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn, TEST_DATA, sheet_name="FirstSheet", table_name="FirstTable") as saver:
            saver.add_sheet(TEST_DATA_WITH_ENUM, sheet_name="AnotherSheet", table_name="AnotherTable")
        verify_using_pyopenxl(fn, sheet_name="FirstSheet", table_name="FirstTable", dimensions="A1:C3", data=[
            ("a", "b", "c"),
            (1, "2", None),
            (None, "B", None),
        ])
        verify_using_pyopenxl(fn, sheet_name="AnotherSheet", table_name="AnotherTable", dimensions="A1:B2", data=[
            ("a", "enum"),
            (1, "ONE"),
        ])

def test_empty_then_two_sheets():
    with tempfile.TemporaryDirectory() as tmpdir:
        fn = Path(tmpdir) / "test.xlsx"
        with save_as_xlsx.SaveAsXlsx(fn) as saver:
            saver.add_sheet(TEST_DATA, sheet_name="FirstSheet", table_name="FirstTable")
            saver.add_sheet(TEST_DATA_WITH_ENUM, sheet_name="AnotherSheet", table_name="AnotherTable")
        verify_using_pyopenxl(fn, sheet_name="FirstSheet", table_name="FirstTable", dimensions="A1:C3", data=[
            ("a", "b", "c"),
            (1, "2", None),
            (None, "B", None),
        ])
        verify_using_pyopenxl(fn, sheet_name="AnotherSheet", table_name="AnotherTable", dimensions="A1:B2", data=[
            ("a", "enum"),
            (1, "ONE"),
        ])



from pathlib import Path
from typing import Any

try:
    import openpyxl
except ImportError:
    openpyxl = None  # type: ignore


def verify_using_pyopenxl(fn: Path | str, dimensions: str, data: list[tuple[Any, ...]] | None = None, sheet_name: str = "Sheet1", table_name: str = "Table1"):
    if openpyxl is None:
        return
    xl = openpyxl.load_workbook(fn)
    sheet = xl[sheet_name]
    if dimensions:
        assert sheet.dimensions == dimensions
    if data is not None:
        assert list(sheet.values) == data
    table = sheet.tables[table_name]
    assert table.name == table_name
    assert table.displayName == table_name
    if data is not None:
        assert table.column_names == list(data[0])
    if dimensions:
        assert table.ref == dimensions
